from .feature_extractor import FeaturesExtractor
from .filter_banks import FilterBanks
from .spectrogram import Spectrogram
from . import mfcc
from . import sigproc
from .get_features import preprocess



import os
import logging
from typing import List
import numpy as np
import random
import tensorflow as tf
from tensorflow import keras
import pandas as pd
from concurrent.futures import ThreadPoolExecutor, wait
import tensorflow.keras.backend as K
from tensorflow.keras.models import Model
from tensorflow.keras.layers import *
import sys

sys.path.append("..")
from deepasr.pipeline import Pipeline
from deepasr.augmentation import Augmentation
from deepasr.decoder import Decoder
from deepasr.features import FeaturesExtractor
from deepasr.vocab import Alphabet
from deepasr.utils import read_audio, save_data

logger = logging.getLogger('asr.pipeline')


class CTCPipeline(Pipeline):
    """
    The pipeline is responsible for connecting a neural network model with
    all non-differential transformations (features extraction or decoding),
    and dependencies. Components are independent.
    """

    def __init__(self,
                 alphabet: Alphabet,
                 features_extractor: FeaturesExtractor,
                 model: keras.Model,
                 optimizer: keras.optimizers.Optimizer,
                 decoder: Decoder,
                 sample_rate: int,
                 mono: True,
                 multi_gpu: bool = True):
        self._alphabet = alphabet
        self._model_cpu = model
        self._optimizer = optimizer
        self._decoder = decoder
        self._features_extractor = features_extractor
        self.sample_rate = sample_rate
        self.mono = mono
        self.multi_gpu = multi_gpu
        self._model = self.distribute_model(model) if multi_gpu else model
        self.temp_model = self._model

    @property
    def alphabet(self) -> Alphabet:
        return self._alphabet

    @property
    def features_extractor(self) -> FeaturesExtractor:
        return self._features_extractor

    @property
    def model(self) -> keras.Model:
        return self._model_cpu

    @property
    def decoder(self) -> Decoder:
        return self._decoder

    def preprocess(self,
                   data: List[np.ndarray],
                   is_extracted: bool,
                   augmentation: Augmentation) -> np.ndarray:
        """ Preprocess batch data to format understandable to a model. """

        if is_extracted:  # then just align features
            features = FeaturesExtractor.align(data)
        else:
            features = self._features_extractor(data)
        features = augmentation(features) if augmentation else features
        # labels = self._alphabet.get_batch_labels(transcripts)
        return features

    def ctc_loss(self, args):
        """ The CTC loss using TensorFlow's `ctc_loss`. """
        y_pred, labels, input_length, label_length = args
        return K.ctc_batch_cost(labels, y_pred, input_length, label_length)

    def compile_model(self, label_dim):
        """ The compiled model means the model configured for training. """

        input_data = self._model.inputs[0]
        y_pred = self._model.outputs[0]

        # your ground truth data. The data you are going to compare with the model's outputs in training
        labels = Input(name='the_labels', shape=[label_dim], dtype='float32')
        # the length (in steps, or chars this case) of each sample (sentence) in the y_pred tensor
        input_length = Input(name='input_length', shape=[1], dtype='float32')
        #  the length (in steps, or chars this case) of each sample (sentence) in the y_true
        label_length = Input(name='label_length', shape=[1], dtype='float32')
        output = Lambda(self.ctc_loss, output_shape=(1,), name='ctc')([y_pred, labels, input_length, label_length])
        self._model = Model(inputs=[input_data, labels, input_length, label_length], outputs=output,
                            name="DeepAsr")
        self._model.compile(loss={'ctc': lambda y_true, y_pred: y_pred}, optimizer=self._optimizer,
                            metrics=['accuracy'])

        self._model.summary()
        logger.info("Model is successfully compiled")

    def fit_iter(self,
                 train_dataset: pd.DataFrame,
                 augmentation: Augmentation = None,
                 prepared_features: bool = False,
                 iter_num: int = 1000,
                 batch_size: int = 32,
                 epochs: int = 3,
                 checkpoint: str = None,
                 **kwargs) -> keras.callbacks.History:
        """ Get ready data, compile and train a model. """

        history = keras.callbacks.History()

        audios = train_dataset['path'].to_list()

        labels = self._alphabet.get_batch_labels(train_dataset['transcripts'].to_list())

        transcripts = train_dataset['transcripts'].to_list()

        train_len_ = len(transcripts)

        if not self._model.optimizer:  # a loss function and an optimizer
            self.compile_model(labels.shape[1])  # have to be set before the training

        for i in range(iter_num):
            train_index = random.sample(range(train_len_ - 25), batch_size)

            x_train = [audios[i] for i in train_index]

            y_train = [labels[i] for i in train_index]

            y_trans = [transcripts[i] for i in train_index]

            train_inputs = self.wrap_preprocess(x_train,
                                                y_train,
                                                y_trans, augmentation, prepared_features)

            outputs = {'ctc': np.zeros([batch_size])}

            # print(train_inputs['the_input'].shape)
            # print(train_inputs['the_labels'].shape)
            # print(train_inputs['input_length'].shape)
            # print(train_inputs['label_length'].shape)
            # print(train_inputs['input_length'])
            # print(train_inputs['label_length'])

            if i % 100 == 0:
                print("iter:", i)
                print("input features: ", train_inputs['the_input'].shape)
                print("input labels: ", train_inputs['the_labels'].shape)
                history = self._model.fit(train_inputs, outputs,
                                          batch_size=batch_size,
                                          epochs=epochs,
                                          verbose=1, **kwargs)
                if checkpoint:
                    self.save(checkpoint)
                    print("Pipeline Saved at ", checkpoint)
            else:
                history = self._model.fit(train_inputs, outputs,
                                          batch_size=batch_size,
                                          epochs=epochs,
                                          verbose=0, **kwargs)

        return history

    def fit(self,
            train_dataset: pd.DataFrame,
            augmentation: Augmentation = None,
            prepared_features: bool = False,
            batch_size: int = 32,
            epochs: int = 3,
            checkpoint: str = None,
            **kwargs) -> keras.callbacks.History:
        """ Get ready data, compile and train a model. """

        audios = train_dataset['path'].to_list()

        labels = self._alphabet.get_batch_labels(train_dataset['transcripts'].to_list())

        transcripts = train_dataset['transcripts'].to_list()

        if not self._model.optimizer:  # a loss function and an optimizer
            self.compile_model(labels.shape[1])  # have to be set before the training

        train_inputs = self.wrap_preprocess(audios,
                                            list(labels),
                                            transcripts, augmentation, prepared_features)

        outputs = {'ctc': np.zeros([batch_size])}

        print("input features: ", train_inputs['the_input'].shape)
        print("input labels: ", train_inputs['the_labels'].shape)

        history = self._model.fit(train_inputs, outputs,
                                  batch_size=batch_size,
                                  epochs=epochs,
                                  verbose=1, **kwargs)

        return history

    def fit_generator(self, train_dataset: pd.DataFrame,
                      shuffle: bool = True,
                      augmentation: Augmentation = None,
                      prepared_features: bool = False,
                      batch_size: int = 32,
                      epochs: int = 3,
                      verbose: int = 1,
                      **kwargs) -> keras.callbacks.History:

        """ Get ready data, compile and train a model. """

        audios = train_dataset['path'].to_list()

        labels = self._alphabet.get_batch_labels(train_dataset['transcripts'].to_list())

        transcripts = train_dataset['transcripts'].to_list()

        train_len_ = len(transcripts)

        if not self._model.optimizer:  # a loss function and an optimizer
            self.compile_model(labels.shape[1])  # have to be set before the training

        train_gen = self.get_generator(audios, labels, transcripts,
                                       batch_size, shuffle, augmentation, prepared_features)

        return self._model.fit(train_gen, epochs=epochs,
                               steps_per_epoch=train_len_ // batch_size, verbose=verbose, **kwargs)

    def get_generator(self, audio_paths: List[str], texts: np.array, transcripts: List[str], batch_size: int = 32,
                      shuffle: bool = True, augmentation: Augmentation = None,
                      prepared_features: bool = False):
        """ Data Generator """

        def generator():
            num_samples = len(audio_paths)
            while True:
                x = list()
                y = list()
                if shuffle:
                    temp = list(zip(audio_paths, texts))
                    random.Random(123).shuffle(temp)
                    x, y = list(zip(*temp))

                pool = ThreadPoolExecutor(1)  # Run a single I/O thread in parallel
                future = pool.submit(self.wrap_preprocess,
                                     x[:batch_size],
                                     y[:batch_size], transcripts[:batch_size], augmentation, prepared_features)
                for offset in range(batch_size, num_samples, batch_size):
                    wait([future])
                    batch = future.result()
                    future = pool.submit(self.wrap_preprocess,
                                         x[offset: offset + batch_size],
                                         y[offset: offset + batch_size], transcripts[offset:offset + batch_size],
                                         augmentation, prepared_features)
                    yield batch, {'ctc': np.zeros([batch_size])}

        return generator()

    def wrap_preprocess(self, audios: List[str], the_labels: List[np.array], transcripts: List[str],
                        augmentation: Augmentation = None,
                        prepared_features: bool = False):
        """ Build training data """
        # the_input = np.array(the_input) / 100
        # the_input = x3/np.max(the_input)

        mid_features = [read_audio(audio, sample_rate=self.sample_rate, mono=self.mono) for audio in audios]

        the_input = self.preprocess(mid_features, prepared_features, augmentation)

        the_labels = np.array(the_labels)

        label_len = [len(trans) for trans in transcripts]  # length of each transcription
        label_lengths = np.array(label_len).reshape(-1, 1)  # reshape to 1d

        input_lengths = np.ones((the_labels.shape[0], 1)) * the_labels.shape[1]
        for i in range(the_input.shape[0]):
            input_lengths[i] = the_labels.shape[1]  # num of features from labels

        return {
            'the_input': the_input,
            'the_labels': the_labels,
            'input_length': np.asarray(input_lengths),
            'label_length': np.asarray(label_lengths)
        }

    def predict(self, audio: str, **kwargs) -> List[str]:
        """ Get ready features, and make a prediction. """
        in_features = self._features_extractor([read_audio(audio, sample_rate=self.sample_rate, mono=self.mono)])
        batch_logits = self._model.predict(in_features, **kwargs)
        decoded_labels = self._decoder(batch_logits)
        predictions = self._alphabet.get_batch_transcripts(decoded_labels)
        return predictions

    def save(self, directory: str):
        """ Save each component of the CTC pipeline. """
        # self._model.save(os.path.join(directory, 'model.h5'))
        self.temp_model.save(os.path.join(directory, 'network.h5'))
        self._model.save_weights(os.path.join(directory, 'model_weights.h5'))
        save_data(self._optimizer, os.path.join(directory, 'optimizer.bin'))
        save_data(self._alphabet, os.path.join(directory, 'alphabet.bin'))
        save_data(self._decoder, os.path.join(directory, 'decoder.bin'))
        save_data(self.multi_gpu, os.path.join(directory, 'multi_gpu_flag.bin'))
        save_data(self.sample_rate, os.path.join(directory, 'sample_rate.bin'))
        save_data(self._features_extractor,
                  os.path.join(directory, 'feature_extractor.bin'))

    # def load(self, directory: str):
    #     """ Load each component of the CTC pipeline. """
    #     # model = keras.models.load_model(os.path.join(directory, 'model.h5'),
    #     #                                 custom_objects={'clipped_relu': cls.clipped_relu})
    #     self._model.load_weights(os.path.join(directory, 'model_weights.h5'))
    #     self._alphabet = load_data(os.path.join(directory, 'alphabet.bin'))
    #     self._decoder = load_data(os.path.join(directory, 'decoder.bin'))
    #     self._features_extractor = load_data(
    #         os.path.join(directory, 'feature_extractor.bin'))

    @staticmethod
    def distribute_model(model: keras.Model) -> keras.Model:
        """ Replicates a model on different GPUs. """
        try:
            strategy = tf.distribute.experimental.MultiWorkerMirroredStrategy()
            with strategy.scope():
                dist_model = model
            print("Training using multiple GPUs")
            logger.info("Training using multiple GPUs")
        except ValueError:
            dist_model = model
            logger.info("Training using single GPU or CPU")
        return dist_model



import os
from typing import List
import numpy as np


class Alphabet:
    """
    Read alphabet-pl.txt, which is the list of valid characters. Alphabet has
    two special characters:
    - space on the beginning
    - blank token default added as the last char

    This class is used to convert characters to labels and vice versa.
    """

    def __init__(self, file_path: str = None, lang: str = None):
        self.size = 0
        self.blank_token = None
        self._label_to_str = []
        self._str_to_label = {}
        if lang in ['en', 'pl']:
            directory = os.path.dirname(os.path.abspath(__file__))
            file_path = os.path.join(directory, f'alphabet-{lang}.txt')
        elif not file_path:
            raise ValueError
        self.process_alphabet_file(file_path)

    def __contains__(self, char: str) -> bool:
        """ Check if char is in the Alphabet. """
        return char in self._str_to_label

    def string_from_label(self, label: int) -> str:
        """ Convert label to string. """
        return self._label_to_str[label]

    def label_from_string(self, string: str) -> int:
        """ Convert string to label. """
        return self._str_to_label[string]

    def process_alphabet_file(self, file_path: str):
        """ Read alphabet-pl.txt file. """
        with open(file_path) as file:
            for line in file:
                if line.startswith('#'):
                    continue
                # Char can contain more than one letter
                char = line[:-1]  # remove the line ending
                self._label_to_str.append(char)
                self._str_to_label[char] = self.size
                self.size += 1
            # Blank token is added on the end
            self.blank_token = self.size - 1

    def get_batch_labels(self, transcripts: List[str]) -> np.ndarray:
        """ Convert batch transcripts to labels """
        batch_labels = [[self.label_from_string(c) for c in transcript.lower().strip() if c in self]
                        for transcript in transcripts]
        max_len = max(map(len, batch_labels))
        default_value = self.blank_token
        for labels in batch_labels:
            remainder = [default_value] * (max_len - len(labels))
            labels.extend(remainder)
        return np.array(batch_labels)

    def get_batch_transcripts(self, sequences: List[np.ndarray]) -> List[str]:
        """ Convert label sequences to transcripts. The `-1` also means the
        blank tag """
        return [''.join(self.string_from_label(char_label)
                        for char_label in sequence
                        if char_label not in (-1, self.blank_token))
                for sequence in sequences]


if __name__ == "__main__":
    al = Alphabet(lang='en')
    labels = al.get_batch_labels(["Hi how are you", "i am vey well, what about you"])
    print(labels)



import abc
from typing import List
import numpy as np


class FeaturesExtractor:

    def __call__(self, batch_audio: List[np.ndarray]) -> np.ndarray:
        """ Extract features from the file list. """
        features = [self.make_features(audio) for audio in batch_audio]
        X = self.align(features)
        return X.astype(np.float16)

    @abc.abstractmethod
    def make_features(self, audio: np.ndarray) -> np.ndarray:
        pass

    @staticmethod
    def standardize(features: np.ndarray) -> np.ndarray:
        """ Standardize globally, independently of features. """
        mean = np.mean(features)
        std = np.std(features)
        return (features - mean) / std

    @staticmethod
    def normalize(audio: np.ndarray):
        """ Normalize float32 signal to [-1, 1] range. """
        gain = 1.0 / (np.max(np.abs(audio)) + 1e-5)
        return audio * gain

    @staticmethod
    def align(arrays: list, default=0) -> np.ndarray:
        """ Pad arrays (default along time dimensions). Return the single
        array (batch_size, time, features). """
        max_array = max(arrays, key=len)
        X = np.full(shape=[len(arrays), *max_array.shape],
                    fill_value=default, dtype=float)
        for index, array in enumerate(arrays):
            time_dim, features_dim = array.shape
            X[index, :time_dim] = array
        return X




import os
import pickle
import logging
from functools import reduce
from logging import Logger
from typing import Any
import numpy as np
import librosa
# from scipy.io import wavfile
from tensorflow import keras

# from google.cloud import storage

logger = logging.getLogger('asr.utils')


def load_data(file_path: str):
    """ Load arbitrary python objects from the pickled file. """
    with open(file_path, mode='rb') as file:
        return pickle.load(file)


def save_data(data: Any, file_path: str):
    """ Save arbitrary python objects in the pickled file. """
    with open(file_path, mode='wb') as file:
        pickle.dump(data, file)


# def download_from_bucket(bucket_name: str, remote_path: str, local_path: str):
#     """ Download the file from the public bucket. """
#     client = storage.Client.create_anonymous_client()
#     bucket = client.bucket(bucket_name)
#     blob = storage.Blob(remote_path, bucket)
#     blob.download_to_filename(local_path, client=client)


# def maybe_download_from_bucket(bucket_name: str, remote_path: str, local_path: str):
#     """ Download file from the bucket if it does not exist. """
#     if os.path.isfile(local_path):
#         return
#     directory = os.path.dirname(local_path)
#     os.makedirs(directory, exist_ok=True)
#     logger.info('Downloading file from the bucket...')
#     download_from_bucket(bucket_name, remote_path, local_path)


def read_audio(file_path: str, sample_rate: int, mono: bool) -> np.ndarray:
    """ Read already prepared features from the store. """
    # fs, audio = wavfile.read(file_path)
    audio = librosa.core.load(file_path, sr=sample_rate, mono=mono)[0]
    return audio


def calculate_units(model: keras.Model) -> int:
    """ Calculate number of the model parameters. """
    units = 0
    for parameters in model.get_weights():
        units += reduce(lambda x, y: x * y, parameters.shape)
    return units


def create_logger(file_path=None, level=20, name='asr') -> Logger:
    """ Create the logger and handlers both console and file. """
    logger = logging.getLogger(name)
    logger.setLevel(level)
    formatter = logging.Formatter(
        '%(asctime)s [%(levelname)-8s] [%(name)-20s] %(message)s',
        datefmt='%Y-%m-%d %H:%M:%S'
    )
    console = logging.StreamHandler()
    console.setFormatter(formatter)
    logger.addHandler(console)  # handle all messages from logger
    if file_path:
        file_handler = logging.FileHandler(file_path, mode='w')
        file_handler.setFormatter(formatter)
        logger.addHandler(file_handler)
    return logger



import setuptools

with open('README.md') as f:
    long_description = f.read()

setuptools.setup(
    name="deepasr",
    version="0.0.6",
    author="Sai Kumar Yava",
    author_email="saikumar.geek@gmail.com",
    description="Keras(Tensorflow) implementations of Automatic Speech Recognition",
    long_description=long_description,
    long_description_content_type='text/markdown',
    url="https://github.com/scionoftech/DeepAsr",
    include_package_data=True,
    packages=['deepasr'],
    keywords=['deepspeech', 'asr', 'speech recognition', 'speech to text'],
    license='GNU',
    install_requires=['tensorflow>=2.0', 'pandas', 'tables', 'scipy'],
    python_requires='>=3.6',
)



from . import augmentation
from . import decoder
from . import evaluate
from . import features
from . import model
from . import pipeline
from . import utils
from . import vocab

# Version of the deepasr package
__version__ = "0.0.6"
